from lshashing import LSHRandom
import math

### Measure Performance
print("lshashing module")
def get_knn_naive(points, query_point, k):
    answer = []
    for i in points:
        x = (math.sqrt(sum([(a - b) ** 2 for a, b in zip(i, query_point)])), i)
        answer.append(x)
    return sorted(answer, key = lambda x: x[0])[:k]

import numpy as np
sample_data = np.random.randint(size = (20000, 15000), low = 0, high = 100)
print("Sample data shape: ", sample_data.shape, "\n")
point = np.random.randint(size = (1, 15000), low = 0, high = 100)[0]
print("query point")
print(point.shape, "\n")
from time import time
print("Start comparison in searching for 4 NNs")
print("##### search knn traditionaly")
start = time()
k = 4
naive_knn = get_knn_naive(sample_data, point, k)
print("time to perform: ", time() - start, "\n")

print("##### Search with lshashing package:")
start = time()
lsh_random = LSHRandom(sample_data, 10, num_tables = 1)
print("time to construct lsh: ", time() - start)
start = time()
nns = lsh_random.knn_search(sample_data, point, k, 5, radius = 2)
print("time to perform: ", time() - start, "\n")

print("##### Search with lshashing package in parallel:")
start = time()
lsh_random2 = LSHRandom(sample_data, 10, num_tables = 1, parallel = True)
print("time to construct lsh: ", time() - start)
start = time()
nns2 = lsh_random.knn_search(sample_data, point, k, 5, radius = 2, parallel = True)
print("time to perform: ", time() - start, "\n")

print("##### Now with Scikit Learn")
from sklearn.neighbors import NearestNeighbors
from sklearn.neighbors import KDTree as kdt

start = time()
nbrs = NearestNeighbors(n_neighbors = k, algorithm = 'ball_tree').fit(sample_data)
print("time to construct ball_tree: ", time() - start)
start = time()
distances, indices = nbrs.kneighbors([point])
print("time to perform: ", time() - start, "\n")
sklearn_knn = sample_data[indices]

print("##### With sklearn KDTree")
start = time()
kdt = kdt(sample_data, metric='euclidean')
print("time to construct the tree: ", time() - start)
start = time()
nearest_dist, nearest_ind = kdt.query([point], k = k)
print("time to perform: ", time() - start, "\n")
kdt_knn = sample_data[nearest_ind[0]]

print("##### basic scikit-learn")
start = time()
nbrs = NearestNeighbors(n_neighbors=k).fit(sample_data)
print("time to fit dataset: ", time() - start)
start = time()
distances, indices = nbrs.kneighbors([point])
print("time to perform: ", time() - start, "\n")
sklearn_knn = sample_data[indices]



from bisect import insort

## to store nearest neighbors
class DistHeap:
    def __init__(self):
        self._container = []

    def push(self, item):
        insort(self._container, item)  # in by sort

    def pushitems(self, items):
        for i in items:
            self.push(i)

    def __getitem__(self, item):
        return self._container[item]

    def __len__(self):
        return len(self._container)

    def __repr__(self):
        return repr(self._container)




from collections import namedtuple
import numpy as np
from itertools import combinations
#import math

NN = namedtuple("NN", "index distance value")

class NNeighbor:
    def __init__(self, *args):
        self.nn = NN(*args)
        self.index = self.nn.index
        self.distance = self.nn.distance
        self.value = self.nn.value

    def __getitem__(self, item):
        return self.nn[item]

    def __lt__(self, other):
        return self.nn.distance < other.nn.distance

    def __repr__(self):
        return f"Neighbor(index={self.nn.index}, distance={self.nn.distance}, value=[{self.nn.value[:3]}...])"

def generate_rand_proj(hash_len, dim):
    return np.random.randn(dim, hash_len)

def unbinarize(binary):
    y = "".join((binary * 1).astype(str))
    return int(y, 2)

def hash_fun(data, rand_proj):
    binaries = np.dot(data, rand_proj) > 0
    return [unbinarize(binary) ** 2 for binary in binaries]

def all_combs(hash_len, radius):
    return list(combinations(range(hash_len), radius))

def visited(visited_bins, key):
    if key in visited_bins:
        return True
    else:
        visited_bins.update([key])
        return False

def euclidean_dist(a, b):
    return np.sqrt(np.sum(np.array([(x - y) ** 2 for x, y in zip(a, b)])))
    #return np.sum(np.array([(x - y) ** 2 for x, y in zip(a, b)]))

def get_distances(candidate, candidate_point, query_point, dist_func):
    return NNeighbor(candidate, dist_func(candidate_point, query_point), candidate_point)



from joblib import Parallel, delayed
from .hash_table import HashTable
from .pq import DistHeap
from .util import NNeighbor, euclidean_dist, get_distances

def parallel_fill_table(data, dims, hash_len, num_tables):
    def _hash_build(data, hash_len, table, dims):
        t = HashTable(hash_len, dims)
        t.build_table(data)
        return t
    tables = Parallel(n_jobs = -2, max_nbytes = None)(delayed(
        _hash_build)(data, hash_len, t, dims) for t in range(num_tables))
    return tables

def parallel_knn_search(tables, data, query_point, k, buckets, dist_func, radius):
    def tables_search(tables, query_point, k, buckets, radius):
        tables_cands = Parallel(n_jobs = -2, max_nbytes = None)(
            delayed(t._knn_search)(query_point, k, buckets, radius) for t in tables)
        return tables_cands
    def cand_distances(cands, data, query_point, dist_func):
        nns = Parallel(n_jobs = -2, max_nbytes = None)(
            delayed(get_distances)(c, data[c, :], query_point, dist_func) for c in cands)
        return nns
    candidates = tables_search(tables, query_point, k, buckets, radius)
    cand_set = set()
    for cand in candidates:
        cand_set.update(cand)
    nns = cand_distances(cand_set, data, query_point, dist_func)
    nn_heap = DistHeap()
    nn_heap.pushitems(nns)
    return nn_heap#[:k]



from .hash_table import HashTable
from .pq import DistHeap
from .util import NNeighbor, euclidean_dist, get_distances
from .para_util import parallel_fill_table, parallel_knn_search

class LSHRandom:
    def __init__(self, data, hash_len, num_tables = 1, parallel = False):
        self.n_rows, self.dims = data.shape
        self.hash_len = hash_len
        #self.tables = []
        if parallel:
            self.tables = parallel_fill_table(data, self.dims, hash_len, num_tables)
            #self.tables.extend(tables)
        else:
            self.tables = []
            for i in range(num_tables):
                t = HashTable(hash_len, self.dims)
                self.tables.append(t)
                self.tables[i].build_table(data)

    def add_new_entry(self, data_point):
        for t in self.tables:
            t.add_new_entry(data_point, self.n_rows)
        self.n_rows += data_point.shape[0]

    def knn_search(self, data, query_point, k, buckets,
                   dist_func = euclidean_dist, radius = 2, parallel = False):
        if parallel:
            nn_heap = parallel_knn_search(self.tables, data, query_point, k, buckets, dist_func, radius)
        else:
            best_candidates = set()
            for t in self.tables:
                tbc = t._knn_search(query_point, k, buckets, radius)
                best_candidates.update(tbc)
            nn_heap = DistHeap()
            for c in best_candidates:
                candidate_point = data[c, :]
                nn = get_distances(c, candidate_point, query_point, dist_func)
                nn_heap.push(nn)
        return nn_heap[:k]






from lshashing import LSHRandom
import numpy as np

def test_lshashing():
    sample_data = np.random.randint(size = (20, 20), low = 0, high = 100)
    point = np.random.randint(size = (1, 20), low = 0, high = 100)
    k = 4
    ntables = 2
    lsh_random = LSHRandom(sample_data, 4, num_tables = 2)
    assert len(lsh_random.tables) == ntables
    nns = lsh_random.knn_search(sample_data, point[0], k, 3)
    assert len(nns) == k

def test_parallel():
    sample_data = np.random.randint(size = (20, 20), low = 0, high = 100)
    point = np.random.randint(size = (1, 20), low = 0, high = 100)
    k = 4
    lsh_random = LSHRandom(sample_data, 4, parallel = True)
    tbl0_elem = [elem for bucket in lsh_random.tables[0].hash_table.values() for elem in bucket]
    assert len(tbl0_elem) == len(sample_data)
    nns = lsh_random.knn_search(sample_data, point[0], 4, 3, parallel = True)
    assert len(nns) == k



from .util import all_combs, visited
from random import shuffle, randint

def fill_table(dct, hashes, n_rows = 0):
    indices = list(range(n_rows, n_rows + len(hashes)))
    for i, h in zip(indices, hashes):
        if h in dct:
            dct[h].append(i)
        else:
            dct[h] = [i]

#def bincount(n):
#    return bin(n).count("1")

def bit_glue(bit, hash_len):
    bit_len = len(bin(bit)[2:])
    z = str(0b0) * (hash_len - bit_len)
    nb = list(z + bin(bit)[2:])
    return nb

def bit_mutate(combs, bit, hash_len):
    nb = bit_glue(bit, hash_len)
    shuffle(combs)
    if len(combs) == 0:
        combs.append((randint(0, hash_len)))
    mutation = combs.pop()
    if isinstance(mutation, int):
        if (len(nb) - 1) < mutation:
            mutation = len(nb) - 1
        nb[mutation] = "0" if nb[mutation] == "1" else "1"
    else:
        for i in mutation:
            nb[i] = "0" if nb[i] == "1" else "1"
    return int("".join(nb), 2)

def search_near_hashes(table, new_bin, bins_to_search, k_searched, k_to_search, radius):
    candidates = set()
    visited_bins = {new_bin}
    hash_len = max(table.keys()).bit_length()
    combs = all_combs(hash_len, radius)
    #tbl_keys = np.array(list(table.keys()))
    #bin_dist = np.array([bincount(new_bin ^ i) for i in tbl_keys])
    #bins = tbl_keys[np.where(bin_dist <= radius)]
    while k_searched < k_to_search or len(visited_bins) < bins_to_search:
        new_bin = bit_mutate(combs, new_bin, hash_len)
        new_bin = new_bin if new_bin in table else min(table.keys(), key = lambda x: abs(x - new_bin))
        if not visited(visited_bins, new_bin):
            candidates.update(table[new_bin])
            k_searched = k_searched + len(candidates)
    return candidates

    